# -*- coding: utf-8 -*-
"""
created 2020/06/15, Markus K\"uhbach, m.kuehbach@mpie.de
Python class to be used in Jupyter notebooks for showing how to run complex workflows of individual analysis with multiple tools
using PARAPROBE. We use both trivial and non-trivial parallelism on a cluster computer to process a large number of reconstructed 
atom probe datasets with existent reconstruction (POS file from IVAS) and ranging (RNG from IVAS) the data here are from
H. Zhao et al. Scripta Mat, 2018, dx.doi.org/10.1016/j.scriptamat.2018.05.024
"""

import sys
#be able to import paraprobe-parmsetup class objects from the 
#directory above this case specific working directory
sys.path.append('../')
import glob
import re
from pathlib import Path
import numpy as np

#import the paraprobe tools we need
from PARAPROBE_BashBatch import *
from PARAPROBE_SlurmSingle import *
from PARAPROBE_BashSingle import *
from PARAPROBE_XML import *
from PARAPROBE_Transcoder import *
#from PARAPROBE_Synthetic import *
from PARAPROBE_Ranging2HDF5 import *
#from PARAPROBE_RNG2RRNG import *
from PARAPROBE_Ranger import *
from PARAPROBE_Surfacer import *
from PARAPROBE_Spatstat import *
from PARAPROBE_Tessellator import *
from PARAPROBE_DBScan import *
#from PARAPROBE_CharParticles import *
#from PARAPROBE_Isosurfacer import *

Compiler = 'ITL'  #last argument is compiler either 'ITL' or 'GCC'

#define the datasets and matching rangefile (reconstructed and with ranging data)
ds_rg = np.array([ ['R76-25946-2h.pos', 'R76-25946-2h.rng'], 
                   ['R76-26865-PA.pos', 'R76-26865-PA.rng'],
                   ['R76-26737-PA.pos', 'R76-26737-PA.rng'],
                   ['R76-26110-30min.pos', 'R76-26110-30min.rng'],
                   ['R76-25139-2h.pos', 'R76-25139-2h.rng'],
                   ['R76-24909-30min.pos', 'R76-24909-30min.rng'],
                   ['R76-26120-OA.pos', 'R76-26120-OA.rng'] ]  )
#ds_rg = np.array([ ['R76-26120-OA.pos', 'R76-26120-OA.rng'] ]  )

#you could add much more datasets here, only your computational resources define the limit...

#transcode existent RNG range file into RRNG format
#for dataset in ds:
#    print('Transcoding RNG range file format to RRNG range file format...')
#    print(str(dataset) + '.rng')
#    task = paraprobe_rng2rrng( dataset + '.rng' )
#    task.write_rrng()
#    rrngfilename = task.get_rrng_filename()

#now set up an ensemble of paraprobe analyses, one analysis, i.e. workflow for each above dataset
#MK::better one workflow per dataset instead of all datasets into the same workflow
#because such we can run all datasets at the same time and in parallel
#so combination of trivial and non-trivial parallelism used the easy way
#this is the same as if we have seven students and give each one dataset and an IVAS license
#and instruct all students to start working. There is a better alternative to this: 

N = np.uint32(1)
AnalysisJobsSlurm = {}
AnalysisJobsBash = {}
for dataset in ds_rg:
    #dataset = ds_rg[0]
    #define a simple numeral a simulation ID, unique for each workflow to
    #SimulationID = N
    #better is to use e.g. five-digit specimen ID and add two-digit recon version
    tmp = re.split('[-|\-v|.]', dataset[0])
    SimulationID = np.uint32(tmp[1])
    
    ProcessesToUse = 1 #use one MPI process to execute, multithreading is on by default#use one MPI process to execute, multithreading is on by default
    
    #define a dictionary of what we need to do for the workflow
    WhatToDo = {}
    
    #zeroth, we transcode IVAS POS to paraprobe HDF5
    task = paraprobe_transcoder( SimulationID, dataset[0] ) #this creates the XML settings file
    ###MK::make changes if desired to the internal XML settings file
    #the numbers in front of the keys words for the WhatToDo dictionaries are necessary!!
    #they assure that the commands in the slurm scripts (slm variable below) end up in order!!
    WhatToDo['0_Transcode'] = task.run( ProcessesToUse )
    #will give me a reconstruction in HDF5 format
    recon_fnm = task.get_recon_filename()
    del task
    
    #first, we transcode the IVAS RNG range file to HDF5
    task = paraprobe_ranging2hdf5( dataset[1] )
    task.write_h5_rangefile()
    #this is done in-place while running this script, the other tools on the cluster
    #WhatToDo['1_Ranging2HDF5']
    del task
    
    #second, we range using our range files
    range_fnm = dataset[1] + '.h5' # + '.rrng'
    task = paraprobe_ranger( SimulationID, recon_fnm, range_fnm )
    WhatToDo['2_Ranger'] = task.run( ProcessesToUse )
    
    #third, we compute an alpha shape to the specimen
    task = paraprobe_surfacer( SimulationID, recon_fnm )
    #make changes to the default settings ##MK::implement getters for the default settings
    #task.set_roi_rmax( 1.0 )
    #task.set_distancing_skin( 1.0 )
    task.set_distancing_complete()
    WhatToDo['3_Surfacer'] = task.run( ProcessesToUse )
    hull_fnm = task.get_hull_filename()
    del task
    
    #fourth, we characterize spatial statistics using the alpha shape and computed distances from the hull file
    task = paraprobe_spatstat( SimulationID, recon_fnm, hull_fnm )
    #these ion combinations e.g. for an Al-Li-Mg-Ag alloy maybe
    task.add_iontype_combi( 'Al', 'Al' )
    task.add_iontype_combi( 'Al', 'Zn' )
    task.add_iontype_combi( 'Al', 'Mg' )
    task.add_iontype_combi( 'Al', 'Cu' )
    task.add_iontype_combi( 'Zn', 'Al' )
    task.add_iontype_combi( 'Zn', 'Zn' )
    task.add_iontype_combi( 'Zn', 'Mg' )
    task.add_iontype_combi( 'Zn', 'Cu' )
    task.add_iontype_combi( 'Mg', 'Al' )
    task.add_iontype_combi( 'Mg', 'Zn' )
    task.add_iontype_combi( 'Mg', 'Mg' )
    task.add_iontype_combi( 'Mg', 'Cu' )
    task.add_iontype_combi( 'Cu', 'Al' )
    task.add_iontype_combi( 'Cu', 'Zn' )
    task.add_iontype_combi( 'Cu', 'Mg' )
    task.add_iontype_combi( 'Cu', 'Cu' )
    #okay, lets do some spatial statistics
    #e.g. radial distribution function
    #task.set_rdf( 0.001, 1.000 )
    task.set_rdf( 0.001, 5.000 )
    #e.g. kth nearest neighbors for all combinations 
    #task.set_knn( 0.001, 1.000, np.array([1]) ) #, 10, 100]) )
    task.set_knn( 0.001, 5.000, np.array([1]) ) #, 10, 100]) )
    #yes a single line gives you all combinations, easy peasy
    #e.g. why not also three-dimensional environment i.e. directional spatial statistics, i.e. 3D spatial distribution maps
    #task.set_sdm( 0.1, 5.0, np.array([1, 10]) )
    WhatToDo['4_Spatstat'] = task.run( ProcessesToUse )
    del task
    
    #lets also build a tessellation of the dataset, that will be useful for the next step...
    task = paraprobe_tessellator( SimulationID, recon_fnm, hull_fnm )
    task.set_cell_erosion( 1.0 )
    WhatToDo['5_Tessellator'] = task.run( ProcessesToUse )
    del task
    
    #sixth, ...because we can now a maximum separation clustering analysis and use
    #the tessellation for marking all cluster with ions closer than here 1.0 nm to the dataset edge as truncated
    #and thereby analyse clusters in the interior or all
    task = paraprobe_dbscan( SimulationID, recon_fnm, hull_fnm )
    #task.set_high_throughput_maximum_separation_method( [0.2, 0.1, 5.00], [5, 10, 5], 1.0 )
    task.set_high_throughput_maximum_separation_method( [0.2, 0.02, 5.00], [5, 10, 5], 1.0 )
    #what is possibly clustering, lets see..., we have a Al-Mg-Zn-Cu alloy, so Mg, Zn, Cu typically form clusters
    task.add_targets( 'Mg' )
    task.add_targets( 'Zn' )
    task.add_targets( 'Cu' )
    #the clou here is that above maximum separation runs will be done for all these...
    WhatToDo['6_DBScan'] = task.run( ProcessesToUse )
    del task
    
    #you could add more analysis steps here, only your imagination is the limit... for example ...
    
# =============================================================================
#     #fifth, we compile a graphical report from the results to assist us in scanning the results of all these analyses, easy peasy
#     #eigth, we run a few isosurface analyses on this (and compute the distances of the ions to the isosurfaces)
#     
#     #for paper20 on HDF5 file format 
#     #ninth, we run clustering methods for other folks in the community for instance Iman Ghamarian and Emmanuel Marquis' Hierarchical DBScan
#     #the benefit of our suggestion to use a common format for presenting results complementary to how individual researchers report their results
#     #is the following:
#     #lets run again our characterize particles tool and compare heads up the two methods
#     
#     #see it has never been so easy to make cross-method comparisons using state-of-the-art scientific computing
#     #we only ask for a minimal amount of agreement how to exchange data and ask people to write a few lines of code
#     #that will boost their professional career, APT data analysis made FAIR and simple!
# =============================================================================

    #now fuse all these steps into a workflow, write a slurm script (to be submitted to workstation or super computer)
    myjobname = str('HZ') + str(SimulationID)
    slm = paraprobe_slurm_single( myjobname[0:9], SimulationID, 'TALOS', ProcessesToUse, WhatToDo, Compiler ) 
    #[0:9] is for the fact that slurm script are abbreviated with first 8 characters
    slm.write_slurm_script()
    #but also write a bash script (to be submitted to a local workstation without super computer)
    bsh = paraprobe_bash_single( myjobname[0:9], SimulationID, 'MPI30', ProcessesToUse, WhatToDo, Compiler )
    bsh.write_bash_script()
 
    
    #remember that we want to submit this job
    AnalysisJobsSlurm[N] = slm.get_slurm_script_filename()
    AnalysisJobsBash[N] = bsh.get_bash_script_filename()
    #increment SimulationID
    N = N + np.uint32(1)

#this last line here makes the submission to the cluster even simpler
#it fuses all above slurm shell scripts into one to submit to cluster at once
sh = paraprobe_bash_batch( 'HuanZhaoScrMatCaseStudy', AnalysisJobsSlurm, AnalysisJobsBash )
#sh.write_bash_script()

#so now just source HuanZhaoScrMatCaseStudy.sh and let the cluster do its work, 100% FAIR

#alternatively, we could also submit the scripts individually
#i.e. sbatch SLURM.PARAPROBE.Workflow.HuanZhaoScrMat2018.SimID.*.sh
#now t